using UnityEngine;
using System.Collections;

public class AIBehavior {
	protected Actor actor;
	public float mTimeout;
	protected float mStartTime;
	
	public AIBehavior(float timeout)
	{
		mTimeout = timeout;
		mStartTime = Time.time;
	}
	
	public AIBehavior()
	{
		mTimeout = 1000000;
		mStartTime = Time.time;
	}
	
	public virtual void SetTimeout(float timeout)
	{
		mTimeout = timeout;
	}
	
	public virtual void SendTimeoutMessage()
	{
		if(actor != null)
			actor.SendMessage("BehaviorTimeout",this);	
	}
	
	public void SetStartTime(float t)
	{
		mStartTime = t;
	}
	
	public void SetActor(Actor a)
	{
		actor = a;
	}
	
	public virtual void GetSteering(ref SteeringOutput output)
	{
		float elapsed = Time.time - mStartTime;
		if(elapsed > mTimeout)
		{
			actor.BehaviorTimeout(this);
			mStartTime = Time.time;
		}
	}
}

public class KinematicSeek : AIBehavior
{
	public override void GetSteering(ref SteeringOutput output)
	{
		base.GetSteering(ref output);
		// First work out the direction
		Transform target = (Transform)actor.GetBlackboard().GetData("SteeringTarget");
        output.linear = target.position;
        output.linear -= (Vector3)actor.GetBlackboard().GetData("Position");

        // If there is no direction, do nothing
        if (output.linear.sqrMagnitude > 0)
        {
            output.linear.Normalize();
            output.linear *= (float)actor.GetBlackboard().GetData("MaxSpeed");
        }
	}
}

public class KinematicPathFollow : AIBehavior
{
	public override void GetSteering(ref SteeringOutput output)
	{
		if (actor.GetDistanceToCurrentPathPoint() < actor.switchPathPointDistance)
		{
			actor.SwitchToNextPathPoint(); 
		}
		
		// First work out the direction
        output.linear = actor.GetCurrentPathPoint() - (Vector3)actor.GetBlackboard().GetData("Position");

        // If there is no direction, do nothing
        if (output.linear.sqrMagnitude > 0)
        {
            output.linear.Normalize();
            output.linear *= (float)actor.GetBlackboard().GetData("MaxSpeed");
        }
	}
}

public class KinematicFlee : AIBehavior
{
	public override void GetSteering(ref SteeringOutput output)
	{
		base.GetSteering(ref output);
		// First work out the direction
		Transform target = (Transform)actor.GetBlackboard().GetData("SteeringTarget");
        output.linear = (Vector3)actor.GetBlackboard().GetData("Position");
        output.linear -= target.position;

        // If there is no direction, do nothing
        if (output.linear.sqrMagnitude > 0)
        {
            output.linear.Normalize();
            output.linear *= (float)actor.GetBlackboard().GetData("MaxSpeed");;
        }
	}
}

public class KinematicArrive : AIBehavior
{
	public float radius = 1.0f;
	public float timeToTarget = 1.0f;
	
	public override void GetSteering(ref SteeringOutput output)
	{
		base.GetSteering(ref output);
		// First work out the direction
        Transform target = (Transform)actor.GetBlackboard().GetData("SteeringTarget");
        output.linear = target.position;
        output.linear -= (Vector3)actor.GetBlackboard().GetData("Position");

        // If there is no direction, do nothing
        if (output.linear.sqrMagnitude < radius*radius)
        {
            output.linear = Vector3.zero;
        }
        else
        {
        	float maxSpeed = (float)actor.GetBlackboard().GetData("MaxSpeed");
            // We'd like to arrive in timeToTarget seconds
            output.linear *= ((float)1.0 / timeToTarget);

            // If that is too fast, then clip the speed
            if (output.linear.sqrMagnitude > maxSpeed*maxSpeed)
            {
                output.linear.Normalize();
                output.linear *= maxSpeed;
            }
        }
	}
}

public class KinematicWander : AIBehavior
{
	
	public KinematicWander(float timeout):base(timeout)
	{
	}
	
	public override void GetSteering(ref SteeringOutput output)
	{
		base.GetSteering(ref output);
		// Move forward in the current direction
		float currentActorOrientation = (float)(actor.GetBlackboard().GetData("Orientation"));
		Vector3 fwd = Actor.GetOrientationAsVector(currentActorOrientation);
		
        output.linear = fwd;
        output.linear *= (float)actor.GetBlackboard().GetData("MaxSpeed");

        // Turn a little
        float change = Random.value - Random.value;
        float maxRotation = (float)actor.GetBlackboard().GetData("MaxRotation");
        output.angular = change * maxRotation;
	}
	
	
	
}

public class KinematicAlign : AIBehavior
{
	public float timeToTarget = 0.1f;
	public override void GetSteering(ref SteeringOutput output)
	{
		base.GetSteering(ref output);
		Actor other = null;
		Transform target = (Transform)actor.GetBlackboard().GetData("SteeringTarget");
		other = (Actor)target.GetComponent("Actor");
		if(other)
		{
			float actorOrientation = (float)(actor.GetBlackboard().GetData("Orientation"));
			float otherOrientation = (float)(other.GetBlackboard().GetData("Orientation"));			
	        float change = (otherOrientation - actorOrientation) % 360.0f;
	        float maxRotation = (float)actor.GetBlackboard().GetData("MaxRotation");
	        output.angular = change * maxRotation;
	        output.angular /= timeToTarget;
		}
	}
}