using UnityEngine;
using System.Collections;

public class Actor : MonoBehaviour{
	//public Transform mSteeringTarget;//TODO: Put this into the blackboard so it can be set and read from other subsystems (i.e. steering and state machines)
	public float mMaxSpeed;
	public float mMaxRotation;
	public bool mFaceWhereYouAreGoing = true;
	public Transform mModelToAnimate;//Visual representation of the Actor in the game.
	public float mPerceptionRange = 100.0f;//Used by the perceptions system
	public float mReactionDistance = 5.0f;//Used by the perception system and flee behavior
	
	public Transform[] pathPoints;
	public float switchPathPointDistance = 0.1f;
	public int currentPathPointIndex = 0;  
	
	private SteeringOutput mSteering = new SteeringOutput(); //TODO: Put into blackboard so other systems can get and set
	private StateMachine mStateMachine;//This is the state machine that will alter the behavior of the actor.  It uses the blackboard to get and set info.
	private Blackboard mBlackboard;//Place to hold perception data and any other stuff that the subsystems need to communicate.
	private PerceptionSystem mPerceptionSystem;//This will query the world for other actors and put their info into the blackboard
	private ArrayList mBehaviors;//This is a list of behaviors that can be changed by the state machine.  Each state can put behaviors that it needs into the list
	
	private float mOrientation;
	private Vector3 mVelocity;
	private Vector3 mLastPosition;
	private Transform mSteeringTarget;

	public Vector3 GetCurrentPathPoint()
	{
		return pathPoints[currentPathPointIndex].position; 
	}
	
	public void SwitchToNextPathPoint()
	{
		currentPathPointIndex += 1;
		if (currentPathPointIndex >= pathPoints.Length)
		{
			currentPathPointIndex = 0;
		}
	}
	
	public float GetDistanceToCurrentPathPoint()
	{
		Vector3 diff = GetCurrentPathPoint() - transform.position;
		return diff.magnitude; 
	}
	
	void Start()
	{
		Game.GetWorld().AddActor(this);//Put yourself into the world so others can see you
		mBehaviors = new ArrayList();
		mBlackboard = new Blackboard();
		mOrientation = gameObject.transform.rotation.eulerAngles.y * Mathf.Deg2Rad;
		//Add properties to the blackboard so other systems can use them (i.e. steering behaviors)
		mBlackboard.SetData("SteeringTarget",mSteeringTarget);
		mBlackboard.SetData("MaxSpeed",mMaxSpeed);
		mBlackboard.SetData("MaxRotation",mMaxRotation);
		mBlackboard.SetData("Orientation",mOrientation);
		mBlackboard.SetData("Position",gameObject.transform.position);
		mBlackboard.SetData("Velocity",mVelocity);
		
		mPerceptionSystem = new PerceptionSystem(Game.GetWorld().GetBlackboard(), mBlackboard,this,mPerceptionRange, 0.0f);
		SetupStateMachine();
		mVelocity = Vector3.zero;
		mModelToAnimate.animation.Play("IdleHold");
	}
	
	void OnDisable()
	{
		Game.GetWorld().RemoveActor(this);
	}
	
	void Update()
	{
		Tick();
	}
	
	public void BehaviorTimeout(AIBehavior beh)
	{
		mStateMachine.SendMessage("MSG_BEHAVIOR_TIMEOUT");	
	}
	
	protected virtual void SetupStateMachine()
	{
		//This can be overrided in inheriting classes
		//Create the state machines and add the states
		mStateMachine = new StateMachine(this);
		StateMachineState idle = new StateIdle(mStateMachine, "IDLE");
		mStateMachine.AddState(idle);
		
		StateMachineState wander = new StateWander(mStateMachine, "WANDER");
		mStateMachine.AddState(wander);
		
		//StateMachineState flee = new StateFlee(mStateMachine, "FLEE");
		//mStateMachine.AddState(flee);
		
		mStateMachine.SetStateByKey("IDLE");//Forcing the initial state to be "IDLE"
	}
	
	public void SetAnimation(string animName)
	{
		//CrossFade is used instead of Play so the animations transition smoothly.
		//Play would cause a very abrupt and jarring change.
		mModelToAnimate.animation.CrossFade(animName);
	}
	
	//Animations can call this to setup a timeout that will send the state machine a message in "seconds" seconds into the future.
	public void AddAnimationTimeout(float seconds)
	{
		//Think of coroutines as threads.  Which means this function call will return immediately.
		StartCoroutine(AnimationTimeout(seconds));
	}
	
	public IEnumerator AnimationTimeout(float seconds)
	{
		yield return new WaitForSeconds(seconds);//This will cause the function to return but then next line will get executed in "seconds" seconds.
		mStateMachine.SendMessage("MSG_ANIMATION_TIMEOUT");	
	}
	
	
	//This is the most important function in the Actor class.
	//This is where all the subsystems get updated and the Actor gets moved and turned.
	//The other subsystems may call Actor functions during their update.  Such as animation playing or setting timeouts.
	public void Tick()
	{
		//Update perception data first.  The other subsystems (i.e. State Machine) will need it.
		mPerceptionSystem.Update();
		Actor nearestActor = (Actor)mBlackboard.GetData("PD.NearestActor");//When getting data from the blackboard you need to cast it to the correct type
		float distToNearestActor = (float)mBlackboard.GetData("PD.DistToNearestActor");

		//Look at perception data and send messages to the state machine if there are other actors near.
		if(nearestActor != null)
		{
			if(distToNearestActor < mReactionDistance)
			{
				mStateMachine.SendMessage("MSG_ACTOR_NEAR");
				//mSteeringTarget = nearestActor.transform;
				mBlackboard.SetData("SteeringTarget", nearestActor.transform);
			}
			else
			{
				mStateMachine.SendMessage("MSG_NO_ACTORS_NEAR");
			}
		}
		
		//Reset steering data and velocity so we don't move unless a behavior tells us to.
		mSteering.Reset();
		mVelocity = Vector3.zero;
		
		//Loop through all of our behaviors.  This list gets changed by the state machine states.  
		//The state machine will modify this list when entering and exiting states.
		for(int i = 0; i < mBehaviors.Count; i++)
		{
			AIBehavior behavior = (AIBehavior)mBehaviors[i];
	
			behavior.GetSteering(ref mSteering);
			mVelocity = mSteering.linear;
			mOrientation += mSteering.angular;
			mOrientation = mOrientation % (2*Mathf.PI);//mOrientation is in radians
		}
		
		//Update mOrientation if the behaviors have set our face where you are going variable.
		if(mFaceWhereYouAreGoing)
			FaceWhereYouAreGoing();
		
		//Move and rotate the actor.
		transform.Translate(mVelocity*Time.deltaTime, Space.World);
		transform.rotation = Quaternion.identity;
		transform.Rotate(0,mOrientation*Mathf.Rad2Deg,0,Space.World);
		
		WrapPositionTomWorld();
		
		//Update the blackboard to have the new values for our data
		mBlackboard.SetData("Orientation",mOrientation);
		mBlackboard.SetData("Position",gameObject.transform.position);
		mBlackboard.SetData("Velocity",mVelocity);
	}
	
	//Makes the world toroidal
	private void WrapPositionTomWorld()
	{
		Vector3 p = transform.position;
		Rect bounds = (Rect)Game.GetWorld().GetBlackboard().GetData("WorldBounds");
		float x = p.x;
		if(x < bounds.xMin)
			x = bounds.xMax - 5;
		else if(x > bounds.xMax)
			x = bounds.xMin + 5;
		
		float z = p.z;
		if(z < bounds.yMin)
			z = bounds.yMax - 5;
		else if(z > bounds.yMax)
			z = bounds.yMin + 5;
			
		transform.position = new Vector3(x, 0.1f,z);	
	}
	
	//Some behaviors won't orient you and will set the mFaceWhereYouAreGoing property instead which will cause this function to be called in Tick()
	private void FaceWhereYouAreGoing()
	{
		if(mVelocity.sqrMagnitude > 0.01f)
		{
			Vector3 movement = mVelocity;
			
			float facingAngle = Mathf.Atan2(movement.x,movement.z);
			
			mOrientation = facingAngle;
			mOrientation = mOrientation % (2*Mathf.PI);
			
		}
		
	}
	
	public static Vector3 GetOrientationAsVector(float mOrientation)
	{
        return new Vector3(Mathf.Sin(mOrientation),
                       0,
                       Mathf.Cos(mOrientation));	
	}
	
	////////////////////////////////////////////////////////////////////////////////////////
	//Getters and Setters
	////////////////////////////////////////////////////////////////////////////////////////
	public GameObject GetGameObject()
	{
		return gameObject;
	}
	
	public Blackboard GetBlackboard()
	{
		return mBlackboard;
	}
	
	public virtual void AddBehavior(AIBehavior beh)
	{
		mBehaviors.Add(beh);
	}
	
	public virtual void ClearBehaviors()
	{
		mBehaviors.Clear();
	}
	
	public virtual void RemoveBehavior(AIBehavior beh)
	{
		mBehaviors.Remove(beh);	
	}
}
